import typing as tp

from satella.coding.recast_exceptions import rethrow_as
from satella.coding.structures import SetHeap


def _is_x_none(x) -> bool:
    return x is None


class merge_series:
    """
    A merger for multiple sequences that return (timestamp, value).

    This will behave as a single-use iterator and return (timestamp, value1, value2, ...)

    :raises ValueError: one of the given series was empty
    """
    __slots__ = ('series', 'next_preloaded_values', 'empty', 'timestamps',
                 'super_next_preloaded_values')

    def __init__(self, *series: tp.Iterator[tp.Tuple[float, tp.Any]]):
        self.series = [iter(x) for x in series]
        try:
            self.next_preloaded_values = [next(x) for x in self.series]
        except StopIteration:
            self.empty = True
            return
        else:
            self.empty = False

        self.timestamps = SetHeap()
        timestamps = {x[0] for x in self.next_preloaded_values}
        max_ts = max(timestamps)
        self.timestamps.push(max_ts)

        self.super_next_preloaded_values = []
        for series in self.series:
            try:
                value = next(series)
                self.super_next_preloaded_values.append(value)
                self.timestamps.push(value[0])
            except StopIteration:
                self.super_next_preloaded_values.append(None)

    def __iter__(self) -> tp.Iterator:
        return self

    def advance(self, i: int) -> None:
        """
        :param i: timestamp to advance to
        :raises ValueError: given series was empty
        """
        if self.super_next_preloaded_values[i] is None:
            raise ValueError('Cannot advance on series %s which is empty' % (i,))
        else:
            self.next_preloaded_values[i] = self.super_next_preloaded_values[i]
            try:
                self.super_next_preloaded_values[i] = next(self.series[i])
                self.timestamps.push(self.super_next_preloaded_values[i][0])
            except StopIteration:
                self.super_next_preloaded_values[i] = None

    def next(self) -> None:
        minimum_timestamp, _ = min(self.next_preloaded_values)
        for i, v in enumerate(self.next_preloaded_values):
            if v[0] == minimum_timestamp:
                self.advance(i)

    def assert_preloaded(self, for_ts: int) -> bool:
        """
        Assert every next preloaded value can at least report for for_ts

        :param for_ts: timestamp to report for
        :return: whether every value can report for for_ts
        """
        for i, _ in enumerate(self.series):
            if self.super_next_preloaded_values[i] is None:
                if self.next_preloaded_values[i][0] > for_ts:
                    return False
            else:
                try:
                    while not (self.next_preloaded_values[i][0] <= for_ts
                               < self.super_next_preloaded_values[i][0]):
                        self.advance(i)
                except TypeError:  # we've reached the finale
                    if self.next_preloaded_values[i][0] > for_ts:
                        return False
        return True

    def assert_have_timestamps(self) -> None:
        """
        Assert that self.timestamps is not empty, or raise StopIteration if it can't be filled in
        """
        if self.timestamps:
            return
        if all(map(_is_x_none, self.super_next_preloaded_values)):
            raise StopIteration('sequence exhausted')
        self.next()
        if not self.timestamps:
            raise StopIteration('cannot advance series anymore')

    @rethrow_as(IndexError, StopIteration)
    def __next__(self):
        if self.empty:
            raise StopIteration('empty right from the start')

        self.assert_have_timestamps()

        ts = self.timestamps.pop()

        while not self.assert_preloaded(ts):
            ts = self.timestamps.pop()  # throws IndexError

        output = (ts, *(v[1] for v in self.next_preloaded_values))
        return output



import inspect

from satella.coding import silence_excs
from satella.coding.structures import Proxy


__all__ = ['Callable']


class Callable(Proxy):
    """
    A wrappable for cases where a Python code expects your function or method
    to have a dict, but it doesn't since Cython compiled them as
    builtin_function_or_method.

    This wrapper wraps your callable nicely and outfits it with a dict.
    """
    def __init__(self, clbl):
        super().__init__(clbl)
        object.__setattr__(self, '__wrapped__', clbl)
        with silence_excs(AttributeError):
            object.__setattr__(self, '__doc__', clbl.__doc__)
        with silence_excs(AttributeError):
            object.__setattr__(self, '__module__', clbl.__module__)
        with silence_excs(AttributeError):
            object.__setattr__(self, '__name__', clbl.__name__)
        with silence_excs(AttributeError):
            object.__setattr__(self, '__annotations__', clbl.__name__)

        try:
            sig = inspect.signature(clbl)
            object.__setattr__(self, '__signature__', sig)
        except (TypeError, ValueError, RecursionError, AttributeError):
            pass



import copy
import importlib

from satella.coding.recast_exceptions import rethrow_as
from satella.configuration import sources
from satella.configuration.sources.base import BaseSource
from satella.exceptions import ConfigurationError

__all__ = [
    'load_source_from_dict',
    'load_source_from_list'
]


def handle_import(dct: dict):
    def convert(v):
        if 'cast_before' in dct:
            v = EXTRA_TYPES[dct['cast_before']['type']](dct['cast_before'])(v)

        return getattr(importlib.import_module(dct['module']), dct['attribute'])(v)

    return convert


EXTRA_TYPES = {
    'binary': lambda dct: dct['value'].encode(dct.get('encoding', 'ascii')),
    'lambda': lambda dct: eval('lambda x: ' + dct['operation'], globals(),
                               locals()),
    'import': handle_import,
}


@rethrow_as(Exception, ConfigurationError)
def load_source_from_dict(dct: dict) -> BaseSource:
    """
    obj has a form of

    {
        "type": "BaseSource",
        "args": []  # optional
        ... kwargs
    }

    :raises ConfigurationError: upon failure to instantiate
    """
    dct = copy.copy(dct)
    type_ = dct.pop('type')  # type: str
    args = dct.pop('args', [])  # type: tp.List
    optional = dct.pop('optional', False)  # type: bool

    def to_arg(arg):
        if isinstance(arg, dict) and 'type' in arg:
            a_type = arg['type']
            if a_type in EXTRA_TYPES:
                return EXTRA_TYPES[a_type](arg)
            elif a_type in sources.__dict__:
                return load_source_from_dict(arg)
            else:
                raise ValueError(
                    'unrecognized argument type %s' % (arg['type'],))
        else:
            return arg

    args = map(to_arg, args)
    kwargs = {k: to_arg(v) for k, v in dct.items()}

    s = sources.__dict__[type_](*args, **kwargs)

    if optional:
        s = sources.OptionalSource(s)
    return s


def load_source_from_list(obj: list) -> 'sources.MergingSource':
    """
    Builds a MergingSource from dict-ed objects
    """
    return sources.MergingSource(*map(load_source_from_dict, obj))



import operator
import typing as tp

from satella.coding.typing import Predicate
from satella.configuration.schema import Descriptor

__all__ = ['x', 'build_structure']

import warnings


def _nop(v: tp.Any) -> tp.Any:
    return v


def make_operation_two_args(operation_two_args: tp.Callable[[tp.Any, tp.Any], tp.Any],
                            docstring: tp.Optional[str] = None,
                            swap_order: bool = False) -> Predicate:
    def operation(self, a: tp.Callable) -> Predicate:
        if isinstance(a, PredicateClass):
            if swap_order:
                def op(v):
                    return operation_two_args(a(v), self(v))

                return PredicateClass(op)
            else:
                def op(v):
                    return operation_two_args(self(v), a(v))

                return PredicateClass(op)
        else:
            if swap_order:
                def op(v):
                    return operation_two_args(a, self(v))

                return PredicateClass(op)
            else:
                def op(v):
                    return operation_two_args(self(v), a)

                return PredicateClass(op)

    if docstring:
        operation.__doc__ = docstring

    return operation


def make_operation_single_arg(operation: tp.Callable[[tp.Any], tp.Any],
                              docstring: tp.Optional[str] = None) -> Predicate:
    def operation_v(self) -> Predicate:
        def operate(v):
            return operation(self.operation(v))

        return PredicateClass(operate)

    if docstring:
        operation_v.__doc__ = docstring

    return operation_v


def _has_keys(a: tp.Container, keys) -> bool:
    for key in keys:
        if key not in a:
            return False
    return True


def _one_of(a, values: tp.Container) -> bool:
    return a in values


class PredicateClass:
    """
    A shorthand to create lambdas using such statements, for example:

    >>> add_two = x + 2
    >>> assert add_two(2) == 4
    """
    __slots__ = ('operation',)

    def __init__(self, operation: tp.Callable[[tp.Any], tp.Any] = _nop):
        self.operation = operation

    str = make_operation_single_arg(str, 'Call str() on predicate')
    int = make_operation_single_arg(int, 'Call int() on predicate')
    float = make_operation_single_arg(float, 'Call float() on predicate')
    false = make_operation_single_arg(lambda y: not y, 'Return a predicate checking whether '
                                                       'value is False')

    def __call__(self, *args) -> bool:
        if len(args) == 0:
            return PredicateClass(lambda y: self.operation(y)())
        else:
            return self.operation(args[0])

    def has_keys(self, *keys) -> Predicate:
        """
        Return a predicate checking whether this value has provided keys
        """
        return make_operation_two_args(_has_keys)(self, keys)

    def one_of(self, *values) -> Predicate:
        """
        Return a predicate checking if x is amongst values
        """
        return make_operation_two_args(_one_of)(self, values)

    def has_p(self, predicate: 'PredicateClass') -> Predicate:
        """
        An old name for has().

        It's deprecated. Use has() instead

        .. deprecated:: 2.14.22
        """
        warnings.warn('This is deprecated and will be removed in Satella 3.x.'
                      'Please use has() instead', DeprecationWarning)
        return self.has(predicate)

    def type(self) -> tp.Type:
        """
        Return a predicate returning the type of it's argument
        """

        def type_of(v):
            return type(self.operation(v))

        return PredicateClass(type_of)

    def is_instance(self, *args):
        """
        Check if given value is one of instances.

        :param args: will be passed as argument to isinstance
        """

        def is_instance(v):
            return isinstance(self.operation(v), args)

        return PredicateClass(is_instance)

    def identity(self) -> Predicate:
        """
        Spawn another object with the same operation, but different identity.

        Used for constructing dicts keyed by predicates.
        """
        return PredicateClass(self.operation)

    def is_valid_schema(self, schema: tp.Optional[tp.Union[Descriptor, tp.Dict]] = None, **kwargs):
        """
        Check if given value has the correct schema.
        The schema is the same as in
        :py:meth:`~satella.coding.structures.DictObject.is_valid_schema`
        """

        def is_schema_correct(v):
            from satella.coding.structures import DictObject
            return DictObject(self.operation(v)).is_valid_schema(schema, **kwargs)

        return PredicateClass(is_schema_correct)

    def has(self, predicate: 'PredicateClass') -> Predicate:
        """
        Check if any element of the current value (which must be an iterable)
        returns True when applied to predicate

        :param predicate: predicate that has to return True for at least one of this predicate's
            values
        """

        def op(v):
            for e in self.operation(v):
                if predicate(e):
                    return True
            return False

        return PredicateClass(op)

    inside = make_operation_two_args(operator.contains,
                                     'Return a predicate checking if x is inside value')

    instanceof = make_operation_two_args(isinstance,
                                         'Return a predicate checking whether this value '
                                         'is an instance of instance')

    length = make_operation_single_arg(len, 'Return a predicate returning length of it''s argument')

    __contains__ = make_operation_two_args(operator.contains)
    __getattr__ = make_operation_two_args(getattr)
    __getitem__ = make_operation_two_args(operator.getitem)
    __eq__ = make_operation_two_args(operator.eq)
    __ne__ = make_operation_two_args(operator.ne)
    __lt__ = make_operation_two_args(operator.lt)
    __gt__ = make_operation_two_args(operator.gt)
    __le__ = make_operation_two_args(operator.le)
    __ge__ = make_operation_two_args(operator.ge)
    __add__ = make_operation_two_args(operator.add)
    __sub__ = make_operation_two_args(operator.sub)
    __mul__ = make_operation_two_args(operator.mul)
    __matmul__ = make_operation_two_args(operator.matmul)
    __and__ = make_operation_two_args(operator.and_)
    __or__ = make_operation_two_args(operator.or_)
    __rshift__ = make_operation_two_args(operator.rshift)
    __lshift__ = make_operation_two_args(operator.lshift)
    __xor__ = make_operation_two_args(operator.xor)
    __truediv__ = make_operation_two_args(operator.__truediv__)
    __floordiv__ = make_operation_two_args(operator.floordiv)
    __mod__ = make_operation_two_args(operator.mod)
    __pow__ = make_operation_two_args(operator.pow)
    __radd__ = make_operation_two_args(operator.add, swap_order=True)
    __rsub__ = make_operation_two_args(operator.sub, swap_order=True)
    __rmul__ = make_operation_two_args(operator.mul, swap_order=True)
    __rmatmul__ = make_operation_two_args(operator.matmul, swap_order=True)
    __rand__ = make_operation_two_args(operator.and_, swap_order=True)
    __ror__ = make_operation_two_args(operator.or_, swap_order=True)
    __rrshift__ = make_operation_two_args(operator.rshift, swap_order=True)
    __rlshift__ = make_operation_two_args(operator.lshift, swap_order=True)
    __rxor__ = make_operation_two_args(operator.xor, swap_order=True)
    __rtruediv__ = make_operation_two_args(operator.__truediv__, swap_order=True)
    __rfloordiv__ = make_operation_two_args(operator.floordiv, swap_order=True)
    __rmod__ = make_operation_two_args(operator.mod, swap_order=True)
    __rpow__ = make_operation_two_args(operator.pow, swap_order=True)
    __neg__ = make_operation_single_arg(operator.neg)
    __invert__ = make_operation_single_arg(operator.invert)
    __abs__ = make_operation_single_arg(abs)

    def __hash__(self) -> int:
        return hash(id(self))


x = PredicateClass()


def build_structure(struct: tp.Union[tuple, list, dict],
                    argument,
                    final_operator=lambda y: y,
                    nested_call=False) -> tp.Union[tuple, list, dict]:
    """
    Given a structure (tuple, list, dict) that contains x's as some of the elements,
    build such a structure corresponding to given that all x's are replaced
    by result of their calculation on argument.

    Just note that if you're constructing dictionaries, use the .identity() method of predicate,
    to randomize it's identity.

    :param struct: structure to build
    :param argument: argument
    :param final_operator: an operator to call on the result
    :param nested_call: internal, don't use
    :return: analogous structure
    """
    if not nested_call:
        v = build_structure(struct, argument, None, True)
        return final_operator(v)

    from satella.coding.structures import HashableWrapper
    if isinstance(struct, tp.MutableMapping):
        new_dict = {}
        for key, value in struct.items():
            key = build_structure(key, argument, None, True)
            value = build_structure(value, argument, None, True)
            new_dict[key] = value
            return new_dict
    elif isinstance(struct, tp.Sequence):
        new_seq = []
        for elem in struct:
            if isinstance(elem, PredicateClass):
                elem = elem(argument)
            else:
                elem = build_structure(elem, argument, None, True)
            new_seq.append(elem)
        return struct.__class__(new_seq)
    elif isinstance(struct, HashableWrapper):
        obj = getattr(struct, '_Proxy__obj')
        return build_structure(obj, argument, None, True)
    elif isinstance(struct, PredicateClass):
        return struct(argument)
    else:
        return struct



from .call_in_future import call_in_future
from .futures import Future, WrappingFuture, InvalidStateError, wrap_if
from .wrapped_executor import ExecutorWrapper

__all__ = ['Future', 'WrappingFuture', 'InvalidStateError', 'call_in_future',
           'ExecutorWrapper', 'wrap_if']



from __future__ import print_function, absolute_import, division

import enum
import logging
import os
import typing as tp
import uuid
from logging import Logger

from satella.coding import silence_excs
from satella.files import DevNullFilelikeObject
from satella.instrumentation import Traceback
from .exception_handlers import BaseExceptionHandler

AsStreamTypeAccept = tp.Union[str, tp.IO, None, Logger, tp.Tuple[Logger, int]]
AsStreamTypeAcceptHR = tp.Union[str, tp.TextIO, Logger, tp.Tuple[Logger, int]]
AsStreamTypeAcceptIN = tp.Union[str, tp.BinaryIO]


class StreamType(enum.IntEnum):
    MODE_FILE = 0  # write to file
    MODE_STREAM = 1  # a file-like object was provided
    MODE_DEVNULL = 2  # just redirect to /dev/null
    MODE_LOGGER = 3   # just a logger


class AsStream:
    __slots__ = ('o', 'human_readable', 'mode', 'file', 'level', 'logger')

    def __init__(self, o: AsStreamTypeAccept, human_readable: bool):
        """
        A stream to dump to

        :param o: stream, or a file name to use, or None to use /dev/null
        :param human_readable: whether the output should be human-readable
            or a pickle (False for pickle)
        """
        self.o = o
        self.human_readable = human_readable

        if isinstance(o, str):
            if os.path.isdir(o):
                self.o = os.path.join(o, uuid.uuid4().hex)

            self.mode = StreamType.MODE_FILE
        elif isinstance(o, tuple) and isinstance(o[0], Logger):
            self.mode = StreamType.MODE_LOGGER
            self.level = o[1]
            self.logger = o[0]
        elif isinstance(o, Logger):
            self.mode = StreamType.MODE_LOGGER
            self.logger = o
            self.level = logging.ERROR
        elif hasattr(o, 'write'):
            self.mode = StreamType.MODE_STREAM

        elif o is None:
            self.mode = StreamType.MODE_DEVNULL
        else:
            raise TypeError('invalid stream object')

    def __enter__(self) -> tp.Union[tp.TextIO, tp.BinaryIO]:
        if self.mode == StreamType.MODE_FILE:
            self.file = open(self.o, 'w' if self.human_readable else 'wb',
                             encoding='utf8' if self.human_readable else None)
            return self.file.__enter__()
        elif self.mode == StreamType.MODE_STREAM:
            return self.o
        elif self.mode == StreamType.MODE_DEVNULL:
            self.o = DevNullFilelikeObject()
            return self.o

    def __exit__(self, exc_type, exc_val, exc_tp):
        if self.mode == StreamType.MODE_FILE:
            return self.file.__exit__(exc_type, exc_val, exc_tp)
        elif self.mode == StreamType.MODE_STREAM:
            with silence_excs(AttributeError):
                self.o.flush()
        elif self.mode == StreamType.MODE_DEVNULL:
            pass


class DumpToFileHandler(BaseExceptionHandler):
    """
    Write the stack trace to a stream-file.

    Note that your file-like objects you throw into that must support only .write() and optionally
    .flush()

    :param human_readables: iterable of either a file-like objects, or paths where
        human-readable files will be output. Also a logger can be put here, or a tuple
        of logger, logging level. Default logging level will be ERROR.
    :param trace_pickles: iterable of either a file-like objects, or paths where pickles with
        stack status will be output
    :raises TypeError: invalid stream
    """
    __slots__ = ('hr', 'tb')

    def __init__(self, human_readables: tp.Iterable[AsStreamTypeAcceptHR],
                 trace_pickles: tp.Iterable[AsStreamTypeAcceptIN] = None):
        super(DumpToFileHandler, self).__init__()
        self.hr = [AsStream(x, True)
                   if not isinstance(x, AsStream) else x
                   for x in human_readables]  # type: tp.List[AsStream]
        self.tb = [AsStream(x, False) if not isinstance(x, AsStream) else x for x in
                   trace_pickles or []]  # type: tp.List[AsStream]

    def handle_exception(self, type_, value, traceback) -> bool:
        try:
            tb = Traceback()
        except ValueError:
            return False  # no traceback, probably hit KeyboardInterrupt or SystemExit,
            # continue with it

        for q in self.hr:
            if q.mode == StreamType.MODE_LOGGER:
                q.logger.log(q.level, str(value), exc_info=value, stack_info=tb.pretty_format())
            else:
                with q as f:
                    f.write('Unhandled exception caught: \n')
                    tb.pretty_print(output=f)

        for q in self.tb:
            with q as f:
                f.write(tb.pickle())



import copy
import sys
import typing as tp
import warnings
from concurrent.futures import Future

from satella.coding.decorators import wraps
from ..cassandra.common import ResponseFuture
from ..cassandra.future import wrap_future

try:
    from opentracing import Span
except ImportError:
    class Span:
        pass


def trace_function(tracer, name: str, tags: tp.Optional[dict] = None,
                   tags_factory: tp.Optional[tp.Union[
                       tp.Dict[str, tp.Callable], tp.List[tp.Tuple[str, tp.Callable]]]] = None):
    """
    Return a decorator that will trace the execution of a given function
    using tracer.start_active_span.

    Can optionally construct it's tags from a predicate building, example:

    >>> class Device:
    >>>     device_id = 'test'
    >>> @trace_function(tracer, 'Processing', tags_factory=[('device_id', x[0].device_id)])
    >>> def process(device: Device):
    >>>     ...

    :param tracer: tracer to use
    :param name: Name of the trace
    :param tags: optional tags to use
    :param tags_factory: a list of tuple (tag name, callable that is called with *args passed to
        this function as a sole argument). Extra tags will be generated from this.
        Can be also a dict.
    """
    if isinstance(tags_factory, dict):
        tags_factory = list(tags_factory.items())

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            nonlocal tags, tags_factory
            my_tags = tags
            if tags_factory is not None:
                if tags is None:
                    tags = {}
                my_tags = copy.copy(tags)
                for key, value in tags_factory:
                    my_tags[key] = value(args)
            with tracer.start_active_span(name, tags=my_tags):
                return fun(*args, **kwargs)

        return inner

    return outer


def trace_future(future: tp.Union[ResponseFuture, Future], span: Span):
    """
    Install a handler that will close a span upon a future completing,
    attaching the exception contents if the future ends with an exception.

    :param future: can be either a normal Future or a Cassandra's ResponseFuture
    :param span: span to close on future's completion
    """
    if isinstance(future, ResponseFuture):
        warnings.warn('Tracing Cassandra futures is deprecated. Use wrap_future() to '
                      'convert it to a standard Python future. This feature will be '
                      'deprecated in Satella 3.x', DeprecationWarning)
        future = wrap_future(future)

    def close_future(fut):
        exc = fut.exception()
        if exc is not None:
            # noinspection PyProtectedMember
            exc_type, value, traceback = sys.exc_info()
            Span._on_error(span, exc_type, value, traceback)
        span.finish()

    future.add_done_callback(close_future)



import json
import typing as tp
from abc import ABCMeta, abstractmethod

import typing

from satella.coding.typing import NoneType

try:
    import ujson
except ImportError:
    pass

__all__ = ['JSONEncoder', 'JSONAble', 'json_encode', 'read_json_from_file',
           'write_json_to_file', 'write_json_to_file_if_different']

Jsonable = tp.TypeVar('Jsonable', list, dict, str, int, float, None)


class JSONAble(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def to_json(self) -> Jsonable:
        """Return a JSON-able representation of this object"""


class JSONEncoder(json.JSONEncoder):
    """
    This encoder will encode everything!
    """

    def default(self, o: tp.Any) -> Jsonable:
        if hasattr(o, 'to_json'):
            return o.to_json()
        elif isinstance(o, (int, float, str, NoneType)):
            return o
        elif isinstance(o, (list, tuple)):
            return [self.default(v) for v in o]
        elif isinstance(o, dict):
            return {self.default(k): self.default(v) for k, v in o.items()}

        try:
            return super().default(o)
        except TypeError:
            dct = {}
            try:
                for k, v in o.__dict__.items():
                    dct[k] = self.default(v)
            except AttributeError:  # o has no attribute '__dict__', try with slots
                try:
                    for slot in o.__slots__:
                        dct[slot] = self.default(getattr(o, slot))
                except AttributeError:      # it doesn't have __slots__ either?
                    return '<an instance of %s>' % (o.__class__.__name__, )
            return dct


def json_encode(x: tp.Any) -> str:
    """
    Convert an object to JSON. Will properly handle subclasses of JSONAble

    :param x: object to convert
    """
    return JSONEncoder().encode(x)


def write_json_to_file(path: str, value: JSONAble) -> None:
    """
    Write out a JSON to a file as UTF-8 encoded plain text.

    :param path: path to the file
    :param value: JSON-able content
    """
    if isinstance(value, JSONAble):
        value = value.to_json()
    with open(path, 'w') as f_out:
        try:
            ujson.dump(value, f_out)
        except NameError:
            json.dump(value, f_out)


def write_json_to_file_if_different(path: str, value: JSONAble) -> bool:
    """
    Read JSON from a file. Write out a JSON to a file if it's value is different,
    as UTF-8 encoded plain text.

    :param path: path to the file
    :param value: JSON-able content
    :return: whether the write actually happened
    """
    if isinstance(value, JSONAble):
        value = value.to_json()
    try:
        val = read_json_from_file(path)
        if val != value:
            write_json_to_file(path, value)
            return True
        return False
    except (OSError, ValueError):
        write_json_to_file(path, value)
        return True


def read_json_from_file(path: str) -> JSONAble:
    """
    Load a JSON from a provided file, as UTF-8 encoded plain text.

    :param path: path to the file
    :return: JSON content
    :raises ValueError: the file contained an invalid JSON
    :raises OSError: the file was not readable or did not exist
    """
    try:
        with open(path, 'r') as f_in:
            return ujson.load(f_in)
    except NameError:
        with open(path, 'r') as f_in:
            try:
                return json.load(f_in)
            except json.decoder.JSONDecodeError as e:
                raise ValueError(str(e))



from setuptools import setup, find_packages
from satella import __version__


setup(keywords=['ha', 'high availability', 'scalable', 'scalability', 'server', 'metrics'],
      packages=find_packages(include=['satella', 'satella.*']),
      version=__version__,
      install_requires=[
            'psutil'
      ],
      test_suite='nose2.collector.collector',
      python_requires='!=2.7.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
      extras_require={
            'HTTPJSONSource': ['requests'],
            'YAMLSource': ['pyyaml'],
            'TOMLSource': ['toml'],
            'FasterJSON': ['ujson'],
            'cassandra': ['cassandra-driver'],
            'opentracing': ['opentracing']
      }
      )
